with Reps, Reps.Ops, Reps.IO, Messages, RG, RG.Beltrami, Ada.Numerics.Generic_Elementary_Functions,Ada.Characters.Latin_1, Round_Up,Intervals, Intervals.Ops, Intervals.IO;
with CoBalls, CoBalls.Ops, CoBalls.IO, Ada.Text_IO,Ada.Float_Text_IO, GetOpt, Ada.Command_Line,Ada.Strings.Unbounded;
use Reps, Reps.Ops, Reps.IO, Messages, Intervals, Intervals.Ops, Intervals.IO,CoBalls, CoBalls.Ops, CoBalls.IO, Ada.Text_IO,Ada.Float_Text_IO, GetOpt, Ada.Command_Line, Ada.Strings.Unbounded;

pragma Elaborate_All(Intervals.Ops);
pragma Elaborate_All(CoBalls.Ops,CoBalls.IO);
pragma Elaborate_All(Reps.Ops);
pragma Elaborate_All (RG);

procedure Form_Grid is

   pragma Suppress(Storage_Check);

   Pdeg:constant Positive:=3000;
   Dho: constant Natural:=3000;
   subtype Scalar is Interval;
   subtype Comp is CoBall;

   package Current_RG is new RG(Pdeg,Dho,Scalar,Comp);
   use Current_RG,Current_RG.RG_Tay,Current_RG.RG_Tay_Ops,Current_RG.RG_Tay_IO,Current_RG.RG_Lin;

   package Current_Beltrami is new Current_RG.Beltrami;
   use Current_Beltrami;

   package Txt_Int_IO is new Ada.Text_IO.Integer_IO(Integer);
   use Txt_Int_IO;

   Test_String : String := "a:n:m:o:r:i:j:z:c:k:y:t:w:g:f:l:";
   ActR,R,R0,Fact,Inc1,Inc2,Inc3,Inc4,Inc5: Rep;
   Sw,J,MaxN,N,M,Va,K: Integer;
   Ga: Scalar;
   Optchar : character;
   Out_Dir,Tay_File: Unbounded_String;
   ShowErr: constant Boolean:=True;
   GetPoly: constant Boolean:=True;
   Decimal: constant Boolean:=False;
begin
   --DEAFULT OPTIONS
   N:=100000;
   M:=65;
   Out_Dir:=To_Unbounded_String("/users/denis/Output/example/");
   Tay_File:=To_Unbounded_String("/users/denis/Output/example/");
   R0:=Rep(0.01);
   Inc1:=Rep(1.0E-6);
   Inc2:=Rep(1.000001);
   Inc3:=Rep(1.0E-3);
   Rho:=Two;
   K:=3;

   --OPTIONS
   Getopt.Opterr := 1;
   loop
      Va := Getopt.Getopt( Test_String );
      exit when Va = -1;
      optchar :=  Character'Val( Va );
      case optchar is
         when 'o' =>
            Out_Dir:=To_Unbounded_String(To_String(Getopt.Optarg));
         when 't' =>
            Tay_File:=To_Unbounded_String(To_String(Getopt.Optarg));
         when 'k' =>
            K:=Integer'Value(To_String(Getopt.Optarg));
         when 'n' =>
            MaxN:=Integer'Value(To_String(Getopt.Optarg));
         when 'm' =>
            M:=Integer'Value(To_String(Getopt.Optarg));
         when 'i' =>
            R:=Rep'Value(To_String(Getopt.Optarg));
         when 'g' =>
            Ga:=Scal(Rep'Value(To_String(Getopt.Optarg)));
         when 'a' =>
            ActR:=Rep'Value(To_String(Getopt.Optarg));
         when 'z' =>
            R0:=Rep'Value(To_String(Getopt.Optarg));
         when 'f' =>
            Fact:=Rep'Value(To_String(Getopt.Optarg));
         when 'c' =>
            Inc1:=Rep'Value(To_String(Getopt.Optarg));
         when 'j' =>
            Inc2:=Rep'Value(To_String(Getopt.Optarg));
         when 'y' =>
            Inc3:=Rep'Value(To_String(Getopt.Optarg));
         when 'w' =>
            Inc4:=Rep'Value(To_String(Getopt.Optarg));
         when 'l' =>
            Inc5:=Rep'Value(To_String(Getopt.Optarg));
         when others =>
            New_Line;
            Message("wrong option in beltrami_solver",Parameter_Error);
      end case;
   end loop;

   declare
      Dir: String:=To_String(Out_Dir);
      Radial_Grid: String:=Dir & "rads/radii";
      VarRho_File: String:=Dir & "rads/varrho";
      RadScal_Grid: String:=Dir & "rads/thick_radii";
      Rho_Grid: String:=Dir & "rads/rhos";
      Periodic_File: String:=Dir & "points/periodic_point";
      ZeroValue_File: String:=Dir & "points/zero_value";
      InfValue_File: String:=Dir & "points/inf_value";
      P_File: String:=Dir & "points/periodic_point";
      Crit_File: String:=Dir & "points/crit_point";
      Cn2_File: String:=Dir & "points/cn2_point";
      Cn2n_File: String:=Dir & "points/cn2n_point";
      RR,Eps: Rep;
      VarRho,VarRhoo,EsSup: Scalar;
      InfValue,ZeroValue,CrPointt: Comp;
      F,F1,F2: File_Type;
      Tay,IterTay: Taylor;
   begin
      Qnmin1:=0;
      Qn:=1;
      Qnplus1:=2;
      Qnplus2:=3;
      for I in 2..K loop
         if I=2 then
            Qnmin2:=0;
            Qnmin1:=1;
            Qn:=2;
            Qnplus1:=3;
            Qnplus2:=5;
         else
            Qnmin2:=Qnmin1;
            Qnmin1:=Qn;
            Qn:=Qnmin2+Qnmin1;
            Qnplus1:=Qn+Qnmin1;
            Qnplus2:=Qnplus1+Qn;
         end if;
      end loop;

      --READING TAYLOR
      New_Line;
      Put("Reading the fixed point ...");
      Read(To_String(Tay_File),Rho,Tay,GetPoly,Decimal);
      Put("done");

      Copy(Tay,IterTay);
      for I in 1..K-1 loop
         IterTay:=Compose(IterTay,Tay);
      end loop;


      New_Line;
      Put("The Taylor series:");
      New_Line;
      Show(Tay);
      New_Line;

      Put("Reading and computing points ...");
      Read(Periodic_File,P,Decimal,ShowErr);
      FixedPoint(Qn,Rho,P,Tay);
      CrPointt:=CoOne;
      --UNCOMMENT IF WANT TO RECOMPUTE THESE POINTS
      --Der(Rho,Rep(0.95)*Rho,IterTay,DerIterTay);
      --Newton(1,Rho,CrPointt,DerIterTay,CoZero);
      Cn2:=Val(Qnplus2,Rho,CrPointt,Tay);
      Cn2n:=Val(Qn,Rho,Cn2,Tay);
      LogB:=Neg(CoI)*Log(CoOne-P/Cn2);--image of C_{n+2} under inverse tau
      LogA:=AbsVal( Log(  (Cn2n-P)*Cn2/((Cn2-P)*Cn2n)  )  );
      LogBB:=LogB;
      Put(" done");

      New_Line;
      Put("Writing points...");
      Write(P_File,P,Decimal,ShowErr);
      Write(Crit_File,CrPointt,Decimal,ShowErr);
      Write(Cn2_File,Cn2,Decimal,ShowErr);
      Write(Cn2n_File,Cn2n,Decimal,ShowErr);
      Put("done");


      New_Line;
      Put("c=");Show(CrPointt);
      New_Line;
      Put("p=");Show(P);
      New_Line;
      Put("Doing InfBelt...");
      InfBelt(Scal(R),P,Tay,InfValue);
      Put("done");

      New_Line;
      Put("Writing the value of the Beltrami differential at infinity...");
      Write(InfValue_File,InfValue,Decimal,ShowErr);
      Put("done");
      New_Line;
      Put("Doing ZeroBelt...");
      ZeroBelt(R0,Ga,Tay,ZeroValue,VarRhoo);
      Put("done");
      VarRho:=Fact*VarRhoo;

      New_Line;
      Put("Writing the value of the Beltrami differential at zero...");
      Write(ZeroValue_File,ZeroValue,Decimal,ShowErr);
      Put("done");

      New_Line;
      Put("Writing the minimum parameter value...");
      Create(F,Out_File,VarRho_File);
      Put(F,VarRho);
      Close(F);
      Put("done");
      New_Line;
      New_Line;
      Put("VALUE AT ZERO=");Show(ZeroValue);
      New_Line;
      Put("MINIMUM PARAMETER VALUE=");Show(VarRho);
      New_Line;
      Put("VALUE AT INFINITY=");Show(InfValue);
      New_Line;
      Put("MAXIMUM PARAMETER VALUE=");Show(Scal(R));
      New_Line;
      Put("MAXIMUM ACTIVE PARAMETER VALUE=");Show(Scal(ActR));

      --DEFINING THE RADIAL GRID
      New_Line;
      Put("Maximal increment in the active range=");Txt_Put(Current_Output,Inc1);
      New_Line;
      Put("Increment outside of the active range=");Txt_Put(Current_Output,Inc4);
      Eps:=Inc1*Inf(VarRho);
      RR:=Approx( Scal(Inf(VarRho),Inf(VarRho+Half*Scal(Eps)) )  );
      J:=2;

      while RR<ActR loop
         if Eps<Inc3 then
            Eps:=Inc2*Eps;
         else
            Eps:=Inc3;
         end if;
         RR:=RR+Eps;
         J:=J+1;
      end loop;
      N:=J;
      New_Line;
      Put("Total number of radii is ");ShortPut(J);
      New_Line;
      declare
         N1: constant Integer:=-N;
         N2: constant Integer:=0;
         Rad,ThickRad: ScalArray(N1..N2);
         Ro: ScalArray(N1-1..N2);
      begin

         --CONSTRUCTING AND WRITING THE RADIAL GRID
         New_Line;
         Put("Constructing and writing the grid...");
         ThickRad(Rad'First):=Scal(Zero,Inf(VarRho));
         Eps:=Inc1*Inf(VarRho);
         ThickRad(Rad'First+1):=Scal(Inf(VarRho),Inf(VarRho+Half*Scal(Eps)) );

         Rad(Rad'First):=ScalZero;
         Rad(Rad'First+1):=Scal(Approx( ThickRad(Rad'First+1) ));

         Ro(Ro'First):=ScalZero;
         Ro(Rad'First):=Scal(Sup( ThickRad(Rad'First) )  );
         Ro(Rad'First+1):=Scal(Sup( ThickRad(Rad'First+1) )  );

         Sw:=0;
         for I in Rad'First+2..Rad'Last loop
            if Eps<Inc3 then
               Eps:=Inc2*Eps;
            else
              Eps:=Inc3;
            end if;
            ThickRad(I):=Scal(Sup(ThickRad(I-1)),Sup(ThickRad(I-1))+Eps);
            Rad(I):=Scal(Approx( ThickRad(I) ));
            if Sw=0 and Eps=Inc3 then
               New_Line;
               Put("Uniform grid starts at ");Show(Rad(I));
               Sw:=1;
            end if;
            Ro(I):=Scal(Sup( ThickRad(I) ) );
         end loop;

         Create(F,Out_File,Radial_Grid);
         Create(F1,Out_File,RadScal_Grid);
         Create(F2,Out_File,Rho_Grid);

         Put(F,Rad'First);  Put(F,Rad'Last);  Put(F,M);
         New_Line(F);

         Put(F1,Rad'First);  Put(F1,Rad'Last); Put(F1,M);
         New_Line(F1);

         Put(F2,Ro'First);  Put(F2,Ro'Last);  Put(F2,M);
         New_Line(F2);

         Txt_Put(F2,Approx(Ro(Ro'First)),Decimal);
         New_Line(F2);
         for I in Rad'Range loop
            Txt_Put(F,Approx(Rad(I)),Decimal);
            New_Line(F);
            Put(F1,ThickRad(I),Decimal);
            New_Line(F1);
            if I<=Ro'Last then
               Txt_Put(F2,Approx(Ro(I)),Decimal);
               New_Line(F2);
            end if;
         end loop;
         Close(F2);
         Close(F1);
         Close(F);
      end;
      Put("done");

      New_Line;
      Put("Computing the essential supremum of the Beltrami differential in a neighborhood of zero...");
      BeltramiSup(M,Inf(VarRhoo),Sup(VarRho),Inc5,Rep(1.0E-5)*CoOne,Tay,EsSup);
      Put("done");
      if Sup(EsSup)>SupMod(ZeroValue) then
         New_Line;
         Put("Rewriting the value of the Beltrami differential at zero...");
         Write(ZeroValue_File,Ball0(Sup(EsSup)),Decimal,ShowErr);
         Put("done");
         New_Line;
         Put("VALUE IN A NBHD OF ZERO=");Show(EsSup);
      end if;
      New_Line;
      Put("Computing the essential supremum of the Beltrami differential at infinity...");
      BeltramiSup(M,ActR,R,Inc4,P,Tay,EsSup);
      Put("done");
      if Sup(EsSup)>SupMod(InfValue) then
         New_Line;
         Put("Rewriting the value of the Beltrami differential at infinity...");
         Write(InfValue_File,Ball0(Sup(EsSup)),Decimal,ShowErr);
         Put("done");
         New_Line;
         Put("VALUE IN A NBHD OF INFINITY=");Show(EsSup);
      end if;
   end;
end Form_Grid;



























