with Messages,  Ada.Text_IO,Ada.Integer_Text_IO;
use Messages, Ada.Text_IO,Ada.Integer_Text_IO;

pragma Elaborate_All (Messages);

package body Lin is

   pragma Suppress(Storage_Check);

   function IMin(I1,I2: Integer) return Integer renames Integer'Min;
   function IMax(I1,I2: Integer) return Integer renames Integer'Max;
   pragma Inline (Imin,Imax);
   pragma Suppress(Storage_Check);


   -- SOME LINAR ALGEBRA PROCEDURES
   function "-"(B: CompArray) return CompArray is
      Bt: CompArray(B'Range);
   begin
      for I in B'Range loop
         Bt(I) := Neg(B(I));
      end loop;
      return Bt;
   end "-";

   procedure SetZero(B: out RepArray) is
   begin
      for I in B'Range loop
         B(I) :=Zero;
      end loop;
   end SetZero;

   procedure SetZero(A: out RepMatrix) is
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
             A(I,J) :=Zero;
         end loop;
      end loop;
   end SetZero;

 procedure SetColumn(J: in Integer; V: in RepArray; A: in out RepMatrix) is
   begin
      for I in A'Range(1) loop
         A(I,J) := V(I);
      end loop;
   end SetColumn;

   procedure Show(B: in RepArray)  is
   begin
      for I in B'Range loop
         New_Line;
         Txt_Put(Current_Output,B(I));
      end loop;
   end Show;

   function "*"(A1,A2: RepMatrix) return RepMatrix is
      R: Rep;
      A: RepMatrix(A1'First(1)..A1'Last(1),A1'First(2)..A1'Last(2));
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            R := Zero;
            for K in A2'Range(1) loop
               R := R+A1(I,K)*A2(K,J);
            end loop;
            A(I,J) := R;
         end loop;
      end loop;
      return A;
   end "*";

   function Scal(R: RepArray) return ScalArray is
      S: ScalArray(R'Range);
   begin
      for I in R'Range loop
         S(I):=Scal(R(I));
      end loop;
      return S;
   end Scal;


   function "+"(B1,B2: CompArray) return CompArray is
      B: CompArray(B2'Range);
   begin
      for I in B'Range loop
         B(I) := B2(I)+B1(I);
      end loop;
      return B;
   end "+";


   function "*"(R: Rep; A: RepMatrix) return RepMatrix is
      Ar: RepMatrix(A'First(1)..A'Last(1),A'First(2)..A'Last(2));
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            Ar(I,J) := R*A(I,J);
         end loop;
      end loop;
      return Ar;
   end "*";

   function Transpose(A: RepMatrix) return RepMatrix is
      TA: RepMatrix(A'First(2)..A'Last(2),A'First(1)..A'Last(1));
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            TA(J,I) := A(I,J);
         end loop;
      end loop;
      return TA;
   end Transpose;

   procedure Transpose(A: in out RepMatrix)  is
   begin
       A:=Transpose(A);
   end Transpose;

   function "*"(R: Rep; B: RepArray) return RepArray is
      Br: RepArray(B'Range);
   begin
      for I in B'Range loop
            Br(I) := R*B(I);
      end loop;
      return Br;
   end "*";


   function "*"(C: Comp; B: CompArray) return CompArray is
      A: CompArray(B'Range);
   begin
      for I in B'Range loop
         A(I) := C*B(I);
      end loop;
      return A;
   end "*";


   function "+"(A1,A2: in CompMatrix) return CompMatrix is
      A: CompMatrix(A2'Range(1),A2'Range(2));
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            A(I,J) := A2(I,J)+A1(I,J);
         end loop;
      end loop;
      return A;
   end "+";


   function "*"(B1,B2: RepArray) return RepArray is
      B: RepArray(B2'Range);
   begin
      for I in B2'Range loop
         B(I) := B1(I)*B2(I);
      end loop;
      return B;
   end "*";

   function "*"(B1,B2: RepArray) return Rep is
      R: Rep;
   begin
      R:=Zero;
      for I in B2'Range loop
         R := R+B1(I)*B2(I);
      end loop;
      return R;
   end "*";


   function "*"(A: RepMatrix; B: RepArray) return RepArray is
      R: Rep;
      Br: RepArray(A'Range(1));
   begin
      for I in Br'Range loop
         R := Zero;
         for J in B'Range loop
            R := R+A(I,J)*B(J);
         end loop;
         Br(I) := R;
      end loop;
      return Br;
   end "*";


   function "*"(B: RepArray; A: RepMatrix) return RepArray is
      R: Rep;
      Br: RepArray(A'Range(2));
   begin
      for I in Br'Range loop
         R := Zero;
         for J in B'Range loop
            R := R+B(J)*A(J,I);
         end loop;
         Br(I) := R;
      end loop;
      return Br;
   end "*";


   function "-"(B1,B2: RepArray) return RepArray is
      B: RepArray(B2'Range);
   begin
      for I in B'Range loop
         B(I) := B1(I)-B2(I);
      end loop;
      return B;
   end "-";

   function "+"(B1,B2: RepArray) return RepArray is
      B: RepArray(B2'Range);
   begin
      for I in B'Range loop
         B(I) := B1(I)+B2(I);
      end loop;
      return B;
   end "+";

   function "-"(B1,B2: CompArray) return CompArray is
      B: CompArray(B2'Range);
   begin
      for I in B'Range loop
         B(I) := B2(I)-B1(I);
      end loop;
      return B;
   end "-";


   function "-"(A1,A2: in CompMatrix) return CompMatrix is
      A: CompMatrix(A2'Range(1),A2'Range(2));
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            A(I,J) := A2(I,J)-A1(I,J);
         end loop;
      end loop;
      return A;
   end "-";


   function "-"(A1,A2: in RepMatrix) return RepMatrix is
      A: RepMatrix(A2'Range(1),A2'Range(2));
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            A(I,J) := A1(I,J)-A2(I,J);
         end loop;
      end loop;
      return A;
   end "-";

   function Prod(B0,B1,B2: RepArray) return Rep is
      R: Rep;
   begin
      R := Zero;
      for I in B0'Range loop
         R := R+(B0(I)*B0(I))*B1(I)*B2(I);
      end loop;
      for I in B0'Last+1 .. B2'Last loop
         R := R+B1(I)*B2(I);
      end loop;
      return R;
   end Prod;


   function NormInf(B: RepArray) return Rep is
      R: Rep;
   begin
      R := Zero;
      for I in B'Range loop
         if Abs(B(I))>R then
            R:=Abs(B(I));
         end if;
      end loop;
      return R;
   end NormInf;


   function NormInf(B: CompArray) return Rep is
      R: Rep;
   begin
      R := Zero;
      for I in B'Range loop
         if SupMod(B(I))>R then
            R:=SupMod(B(I));
         end if;
      end loop;
      return R;
   end NormInf;


      -- norms
   function Norm1(B: RepArray) return Rep is
      R: Rep;
   begin
      R := Zero;
      for I in B'Range loop
         R := R+Abs(B(I));
      end loop;
      return R;
   end Norm1;


   function Norm1(A: RepMatrix) return Rep is
      R,Rj: Rep;
   begin
      R := Zero;
      for J in A'Range(1) loop
         Rj := Zero;
         for I in A'Range(2) loop
            Rj := Rj+Abs(A(I,J));
         end loop;
         if Rj>R then
            R := Rj;
         end if;
      end loop;
      return R;
   end Norm1;


   function Norm2(B: RepArray) return Rep is
      R: Rep;
   begin
      R := Zero;
      for I in B'Range loop
         R := R+Abs(B(I)*B(I));
      end loop;
      return Sqrt(R);
   end Norm2;



   procedure SetZero(B: out CompArray) is
   begin
      for I in B'Range loop
         B(I) :=Complecs(Zero,Zero);
      end loop;
   end SetZero;


   procedure SetZero(B: out ScalArray) is
   begin
      for I in B'Range loop
         B(I) :=Scal_Zero;
      end loop;
   end SetZero;


   procedure SetZero(A: out CompMatrix) is
      pragma Suppress(Storage_Check);
   begin
      for I in A'Range(1) loop
         for J in A'Range(2) loop
            A(I,J) := Complecs(Zero,Zero);
         end loop;
      end loop;
   end SetZero;


   procedure  PointsProdAdd(M: in CompMatrix; H: in out CompMatrix)  is
   begin
      for I in M'Range(1) loop
         for J in M'Range(2) loop
            H(I,J):=M(I,J)*H(I,J)+M(I,J);
         end loop;
      end loop;
   end PointsProdAdd;


   procedure  ProdAdd(Pex: in Rep; Ro: in ScalArray; M: in CompMatrix; H: in out CompMatrix; PLpError: out Scalar)  is
      K1: constant Integer:=M'First(2);
      K2: constant Integer:=M'Last(2);
      RoDiff,EssSup,L2Sum: Scalar;
      EE: Comp;
      C: CompArray(H'Range(2));
   begin
      PLpError:=Scal_Zero;
      EssSup:=Scal_Zero;
      RoDiff:=Scal_Pi*(Ro(M'First(1))-Ro(M'First(1)-1))*(Ro(M'First(1))+Ro(M'First(1)-1));
      L2Sum:=Scal_Zero;
      if not Trunc then
         for J in 2*K1..K1-1 loop
            --Computing the J-th coeeficient of xi
            EE:=Co_Zero;
            for L in IMax(K1,J-K2).. IMin(K2,J-K1) loop
               EE:=EE+H(M'First(1),L)*M(M'First(1),J-L);
            end loop;
            EssSup:=EssSup+AbsVal(EE);
            L2Sum:=L2Sum+Sqr(AbsVal(EE));
         end loop;
      end if;

      for J in K1..K2 loop
         C(J):=Co_Zero;
         for K in IMax(J-K2,K1)..IMin(J-K1,K2) loop
            C(J):=C(J)+H(M'First(1),J-K)*M(M'First(1),K);
         end loop;
      end loop;

      if not Trunc  then
         for J in K2+1..2*K2 loop
            --Computing the J-th coeeficient of xi
            EE:=Co_Zero;
            for L in IMax(K1,J-K2).. IMin(K2,J-K1) loop
               EE:=EE+H(M'First(1),L)*M(M'First(1),J-L);
            end loop;
            EssSup:=EssSup+AbsVal(EE);
            L2Sum:=L2Sum+Sqr(AbsVal(EE));
         end loop;
         PLpError:=PLpError+Scal_Pi*Exp((Pex-Two)*Log(Scal(Sup(EssSup))))*RoDiff*L2Sum;
      end if;

      --I>M'First(1)
      for I in M'First(1)+1..M'Last(1) loop
         EssSup:=Scal_Zero;
         RoDiff:=Scal_Pi*(Ro(I)-Ro(I-1))*(Ro(I)+Ro(I-1));
         L2Sum:=Scal_Zero;
         if not Trunc then
            for J in 2*K1..K1-1 loop
               --Computing the J-th coeeficient of xi
               EE:=Co_Zero;
               for L in IMax(K1,J-K2).. IMin(K2,J-K1) loop
                  EE:=EE+H(I,L)*M(I,J-L);
               end loop;
               EssSup:=EssSup+AbsVal(EE);
               L2Sum:=L2Sum+Sqr(AbsVal(EE));
            end loop;
         end if;

         for J in K1..K2 loop
            H(I-1,J):=C(J)+M(I-1,J);
            C(J):=Co_Zero;
            for K in IMax(J-K2,K1)..IMin(J-K1,K2) loop
               C(J):=C(J)+H(I,J-K)*M(I,K);
            end loop;
         end loop;

         if not Trunc  then
            for J in K2+1..2*K2 loop
               --Computing the J-th coeeficient of xi
               EE:=Co_Zero;
               for L in IMax(K1,J-K2).. IMin(K2,J-K1) loop
                  EE:=EE+H(I,L)*M(I,J-L);
               end loop;
               EssSup:=EssSup+AbsVal(EE);
               L2Sum:=L2Sum+Sqr(AbsVal(EE));
            end loop;
            PLpError:=PLpError+Scal_Pi*Exp((Pex-Two)*Log(Scal(Sup(EssSup))))*RoDiff*L2Sum;
         end if;
      end loop;
      for J in K1..K2 loop
         H(H'Last(1),J):=C(J)+M(H'Last(1),J);
      end loop;
   end ProdAdd;


   procedure  ProdAdd(M: in CompMatrix; H: in out CompMatrix)  is
      K1: constant Integer:=M'First(2);
      K2: constant Integer:=M'Last(2);
      C1: CompArray(M'Range(2));
   begin
      SetZero(C1);
      for I in M'Range(1) loop
         for J in K1..K2 loop
            if I>M'First(1) then
               H(I-1,J):=C1(J)+M(I-1,J);
            end if;
            C1(J):=Co_Zero;
            for K in IMax(J-K2,K1)..IMin(J-K1,K2) loop
               C1(J):=C1(J)+H(I,J-K)*M(I,K);
            end loop;
         end loop;
      end loop;
      for J in K1..K2 loop
         H(H'Last(1),J):=C1(J)+M(H'Last(1),J);
      end loop;
   end ProdAdd;


   function RoughLpNorm(Pex: Rep; Ro: ScalArray; B: CompMatrix) return Scalar is
      L2Norm,EssSup,EsSup,L2Sum,RadDiff,LpNorm,LqNorm: Scalar;
      II: Integer;
      L,R: Rep;
      Qex: constant Scalar:=Scal(Pex)/(Pex-One);
   begin
      EsSup:=Scal_Zero;
      L2Norm:=Scal_Zero;
      if (Ro'First+1)/=B'First(1) then
         II:=B'First(1);
      else
         EsSup:=AbsVal(B(B'First(1),0));
         L2Norm:=L2Norm+Scal_Pi*(Ro(B'First(1))-Ro(B'First(1)-1))*(Ro(B'First(1))+Ro(B'First(1)-1))*Sqr(AbsVal(B(B'First(1),0)));
         II:=  B'First(1)+1;
      end if;
      for I in II..B'Last(1) loop
         RadDiff:=Scal_Pi*(Ro(I)-Ro(I-1))*(Ro(I)+Ro(I-1));
         L2Sum:=Scal_Zero;
         EssSup:=Scal_Zero;
         for J in B'Range(2) loop
            EssSup:=EssSup+AbsVal(B(I,J));
            --L2 norm
            L2Sum:=L2Sum+Sqr(AbsVal(B(I,J)));
         end loop;
         L2Norm:=L2Norm+L2Sum*RadDiff;
         if Sup(EssSup)>Sup(EsSup) then
            EsSup:=EssSup;
         end if;
      end loop;
      LpNorm:=Exp((Scal(Pex)-Scal_Two)*Log(Scal(Sup(EsSup)))/Pex)*Exp(Log(Scal(Sup(L2Norm)))/Pex);
      LqNorm:=Exp((Qex-Scal_Two)*Log(Scal(Sup(EsSup)))/Qex)*Exp(Log(Scal(Sup(L2Norm)))/Qex);
      R:=Sup(LpNorm);
      L:=Inf(L2Norm/LqNorm);
      return Scal(L,R);
   end RoughLpNorm;


   function UpperBoundLpNorm(Pex: Rep; Ro: ScalArray; B: CompMatrix) return Scalar is
      L2Norm,EssSupR,EsSupR: Scalar;
      II: Integer;
      L2Sum,RadDiff: Scalar;
      Qex: constant Scalar:=Scal(Pex)/(Pex-One);
   begin
      EsSupR:=Scal_Zero;
      L2Norm:=Scal_Zero;
      if (Ro'First+1)/=B'First(1) then
         II:=B'First(1);
      else
         EsSupR:=AbsVal(B(B'First(1),0));
         L2Norm:=L2Norm+Scal_Pi*(Ro(B'First(1))-Ro(B'First(1)-1))*(Ro(B'First(1))+Ro(B'First(1)-1))*Sqr(AbsVal(B(B'First(1),0)));
         II:=  B'First(1)+1;
      end if;
      for I in II..B'Last(1) loop
         RadDiff:=Scal_Pi*(Ro(I)-Ro(I-1))*(Ro(I)+Ro(I-1));
         L2Sum:=Sqr(AbsVal(B(I,0)));
         EssSupR:=AbsVal(B(I,0));
         for J in 1..B'Last(2) loop
            EssSupR:=EssSupR+AbsVal(B(I,J));
            --L2 norm
            L2Sum:=L2Sum+Sqr(AbsVal(B(I,J)));
         end loop;
         for J in reverse -1..B'First(2) loop
            EssSupR:=EssSupR+AbsVal(B(I,J));
            --L2 norm
            L2Sum:=L2Sum+Sqr(AbsVal(B(I,J)));
         end loop;
         L2Norm:=L2Norm+L2Sum*RadDiff;
         if Sup(EssSupR)>Sup(EsSupR) then
            EsSupR:=EssSupR;
         end if;
      end loop;
      return Exp((Scal(Pex)-Scal_Two)*Log(Sup(EsSupR))/Pex)*Exp(Log(Sup(L2Norm))/Pex);
   end UpperBoundLpNorm;


   function EssSup(B: CompMatrix) return Scalar is
      --MM is the number of angles on which the esssup is evaluated.
      --N2 and N1 are the cut-off for the coefficients in the Fourier series. Absolute values of higher order coeffs are added to the esssup.
      Cex1,Cex2,Ess: Comp;
      EsSup,EsSupp,Res: Scalar;
      Psi: Scalar;
      MM: constant Integer:=5;--fine angle grid
      Del: Scalar:=Scal(Sup(Scal_Pi))/Rep(MM*B'Length(2));
      Ex1,Ex2: CompArray(1..MM*B'Length(2));
      N2: constant Integer:=-5;
      N1: constant Integer:=5;
      Count: Integer;
   begin
      EsSupp:=Scal_Zero;
      Count:=1;
      Psi:=-Scal_Pi;
      for J in B'Range(2) loop
         for L in 1..MM loop
            Ex1(Count):=Exp(Scal(Inf(Psi),Sup(Psi+Del))*Co_I);
            Ex2(Count):=Inv(Ex1(Count));
            Psi:=Psi+Del;
            Count:=Count+1;
         end loop;
      end loop;

      for I in B'Range(1) loop
         EsSup:=Scal_Zero;
         Res:=Scal_Zero;
         for J in B'First(2)..(N1-1) loop
            Res:=Res+AbsVal(B(I,J));
         end loop;
         Count:=1;
         for J in B'Range(2) loop
            for L in 1..MM loop
               Cex1:=Ex1(Count);
               Ess:=B(I,0);
               for K in 1..N2 loop
                  Ess:=Ess+B(I,K)*Cex1;
                  Cex1:=Cex1*Ex1(Count);
               end loop;
               Cex2:=Ex2(Count);
               for K in reverse -1..N1 loop
                  Ess:=Ess+B(I,K)*Cex2;
                  Cex2:=Cex2*Ex2(Count);
               end loop;
               Count:=Count+1;
               if SupMod(Ess)>Sup(EsSup) then
                  EsSup:=AbsVal(Ess);
               end if;
            end loop;
         end loop;
         for J in N2+1..B'Last(2) loop
            Res:=Res+AbsVal(B(I,J));
         end loop;
         if Sup(EsSup+Res)>Sup(EsSupp) then
            EsSupp:=EsSup+Res;
         end if;
      end loop;
      return EsSupp;
   end EssSup;


   procedure EssSup(B: in CompMatrix; E: out ScalArray) is
      --MM is the number of angles on which the esssup is evaluated.
      --N2 and N1 are the cut-off for the coefficients in the Fourier series. Absolute values of higher order coeffs are added to the esssup
      Cex1,Cex2,Ess: Comp;
      EsSup,Res: Scalar;
      Psi: Scalar;
      MM: constant Integer:=100;--fine angle grid
      Del: Scalar:=Scal_Pi/Rep(MM);
      Ex1,Ex2: CompArray(1..MM);
      N2: constant Integer:=10;
      N1: constant Integer:=-10;
      Count: Integer;
   begin
      Count:=1;
      Psi:=-Scal_Pi;
      for L in 1..MM loop
         Ex1(Count):=Exp(Scal(Inf(Psi),Sup(Psi+Del))*Co_I);
         Ex2(Count):=Exp(Neg(Scal(Inf(Psi),Sup(Psi+Del))*Co_I));
         Psi:=Psi+Del;
         Count:=Count+1;
      end loop;
      for I in B'Range(1) loop
         EsSup:=Scal_Zero;
         Res:=Scal_Zero;
         for J in B'First(2)..(N1-1) loop
            Res:=Res+AbsVal(B(I,J));
         end loop;
         Count:=1;
         for L in 1..MM loop
            Cex1:=Ex1(Count);
            Ess:=B(I,0)+Co_One;
            for K in 1..N2 loop
               Ess:=Ess+B(I,K)*Cex1;
               Cex1:=Cex1*Ex1(Count);
            end loop;
            Cex2:=Ex2(Count);
            for K in reverse -1..N1 loop
               Ess:=Ess+B(I,K)*Cex2;
               Cex2:=Cex2*Ex2(Count);
            end loop;
            Count:=Count+1;
            if SupMod(Ess)>Sup(EsSup) then
               EsSup:=AbsVal(Ess);
            end if;
         end loop;
         for J in N2+1..B'Last(2) loop
            Res:=Res+AbsVal(B(I,J));
         end loop;
         E(I):=EsSup+Res;
      end loop;
   end EssSup;


   function  RoughEssSup(B: CompMatrix) return Scalar is
      Ess,EsSup: Scalar;
   begin
      EsSup:=Scal_Zero;
      for I in B'Range(1) loop
         Ess:=Scal_Zero;
         for J in B'Range(2) loop
            Ess:=Ess+AbsVal(B(I,J));
         end loop;
         if Sup(Ess)>Sup(EsSup) then
            EsSup:=Ess;
         end if;
      end loop;
      return EsSup;
   end RoughEssSup;

end Lin;



















