with Ada.Text_IO, Reps, Reps.Ops, Reps.IO;
use Ada.Text_IO, Reps, Reps.Ops, Reps.IO;
with  CoBalls.IO, CoNumbers.IO, Numerics.IO, Intervals.IO;
use  CoBalls.IO, CoNumbers.IO, Numerics.IO, Intervals.IO;
with Round_Up;

pragma Elaborate_All (Ada.Text_IO);

generic

   type Scalar is private;
   type Comp is private;


   --intervals
   with function IsNumeric(S: Scalar) return Boolean is <>;
   with function Scal(K: Integer) return Scalar is <>;
   with function Scal(R: Rep) return Scalar is <>;
   with function Err0(S: Scalar) return Boolean is <>;
   with function Err(S: Scalar) return Rep is <>;
   with function Approx(S: Scalar) return Rep is <>;
   with function "-"(S: Scalar) return Scalar is <>;
   with function Center(S: Scalar) return Scalar is <>;
   with procedure ResetCenter(S: in out Scalar) is <>;
   with function ResetCenter(S: Scalar) return Scalar is <>;

   with function Scal(R1,R2: Rep) return Scalar is <>;
   with function Ball(S: Scalar) return Scalar is <>;
   with function Ball(S: Rep) return Scalar is <>;
   with function Inf(S: Scalar) return Rep is <>;
   with function Sup(S: Scalar) return Rep is <>;
   with function SupAbs(S: Scalar) return Rep is <>;
   with function "<"(S1,S2: Scalar) return Boolean is <>;
   with function IntFloor(S: Scalar) return Integer is <>;
   with function IntCeiling(S: Scalar) return Integer is <>;
   with function "abs"(S: Scalar) return Scalar is <>;
   with function Min(S1,S2: Scalar) return Scalar is <>;
   with function Max(S1,S2: Scalar) return Scalar is <>;
   with function Cap(R: Radius; S: Scalar) return Scalar is <>;
   with function Up(R: Rep; Dummy: Scalar) return Rep is <>;
   with procedure ErrMult(R: in Rep; S: in out Scalar) is <>;

   with function "+"(S,T: Scalar) return Scalar is <>;
   with function "-"(S,T: Scalar) return Scalar is <>;
   with function "*"(R: Rep; S: Scalar) return Scalar is <>;
   with function "*"(S: Scalar; R: Rep) return Scalar is <>;
   with function "*"(S,T: Scalar) return Scalar is <>;
   with function "/"(S: Scalar; R: Rep) return Scalar is <>;
   with function "/"(R: Rep; S: Scalar) return Scalar is <>;
   with function "/"(S,T: Scalar) return Scalar is <>;
   with function Sqr(S: Scalar) return Scalar is <>;
   with function Inv(S: Scalar) return Scalar is <>;
   with function "**"(S: Scalar; I: Integer) return Scalar is <>;
   with function Short_Exp(R: Scalar; Iter: Integer) return Scalar is <>;

   with function Cosh(S: Scalar) return Scalar is <>;
   with function Sinh(S: Scalar) return Scalar is <>;
   with function Sin(S: Scalar) return Scalar is <>;
   with function Cos(S: Scalar) return Scalar is <>;
   with function Exp(S: Scalar) return Scalar is <>;
   with function Log(S: Scalar) return Scalar is <>;
   with function Sqrt(S: Scalar) return Scalar is <>;
   with procedure CosSin(S: in Scalar; Sc,Ss: out Scalar) is <>;

   with procedure Show(S: in Scalar; NewLine: in Boolean := True) is <>;
   with procedure Get(F: in File_Type; S: out Scalar; Decimal: in Boolean := True) is <>;
   with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := True) is <>;

   --balls

   with function RePart(C: Comp) return Scalar is <>;
   with function ImPart(C: Comp) return Scalar is <>;
   with function Conj(C: Comp) return Comp is <>;
   with function SetBall(C: Comp;R: Rep:=Zero) return Comp is <>;
   with function IsCoNumber(S: Comp) return Boolean is <>;
   with function IsBall(S: Comp) return Boolean is <>;
   with function Complecs(Re,Im: Rep) return Comp is <>;
   with procedure ResetCenter(S: in out Comp) is <>;
   with function ResetCenter(S: Comp) return Comp is <>;
   with procedure Split(S: in Comp; Sr,Se: out Comp) is <>;
   with procedure ResetErr(S: in out Comp) is <>;
   with function ResetErr(S: Comp) return Comp is <>;
   with function CenterBall(S: Comp) return Comp is <>;
   with function Ball0(R: Rep) return Comp is <>;
   with function Ball0(R: Comp) return Comp is <>;
   with function ErrPart(S: Comp) return Rep is <>;

   with function InfRe(S: Comp) return Rep is <>;
   with function SupRe(S: Comp) return Rep is <>;
   with function InfIm(S: Comp) return Rep is <>;
   with function SupIm(S: Comp) return Rep is <>;
   with function SupMod(S: Comp) return Rep is <>;
   with function InfMod(S: Comp) return Rep is <>;
   with function AbsBall(S: Comp) return Comp is <>;
   with function AbsVal(S: Comp) return Scalar is <>;
   with function Arg(C: Comp) return Scalar is <>;
   with procedure ErrMult(R: in Rep; S: in out Comp) is <>;
   with function Neg(S: Comp) return Comp is <>;
   with procedure RandomBall(Seed: in out Long_Integer; S: out Comp) is <>;

   with function Contains(S: Comp; C: Comp) return Boolean is <>;
   with function Intersects(S: Comp; T: Comp) return Boolean is <>;
   with function SetBall(R1,R2: Scalar) return Comp is <>;
   with function "+"(S,T: Comp) return Comp is <>;
   with function "-"(S,T: Comp) return Comp is <>;
   with function "*"(R: Rep; S: Comp) return Comp is <>;
   with function "/"(S: Comp; R: Rep) return Comp is <>;
   with function "*"(R: Scalar; S: Comp) return Comp is <>;
   with function "/"(S: Comp; R: Scalar) return Comp is <>;
   with function Sqr(S: Comp) return Comp is <>;
   with function "*"(S,T: Comp) return Comp is <>;
   with function Inv(S: Comp) return Comp is <>;
   with function Long_Inv(S: Comp) return Comp is <>;
   with function "/"(S,T: Comp) return Comp is <>;
   with function "**"(S: Comp; I: Integer) return Comp is <>;
   with function Cap(R: Radius; S: Comp) return Comp  is <>;
   with function MaxMod(S1,S2: Comp) return Comp is <>;

   with function Cosh(S: Comp) return Comp is <>;
   with function Sinh(S: Comp) return Comp is <>;
   with function Exp(S: Comp) return Comp is <>;
   with function EiExp(S1,S2: Comp; NumErr: Rep:=Rep(1.0E-15)) return Comp is <>;
   with function Cos(S: Comp) return Comp is <>;
   with function Sin(S: Comp) return Comp is <>;
   with function Log(S: Comp) return Comp is <>;
   with function Sqrt(S: Comp) return Comp is <>;

   with procedure Show(S: in Comp; Cut: in Rep:=Zero) is <>;
   with procedure Show(Text: in String; S: in Comp; Cut: in Rep:=Zero) is <>;
   with procedure Get(F: in File_Type; S: out Comp;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True) is <>;
   with procedure Put(F: in File_Type; S: in Comp; Decimal: in Boolean := True; ShowErr: in  Boolean:=True) is <>;

package Lin is

   type Pivot is array (Integer range <>) of Integer;

   type RepArray is array (Integer range <>) of Rep;
   type RepMatrix is array (Integer range <>, Integer range <>) of Rep;


   type ScalArray is array (Integer range <>) of Scalar;
   type ScalMatrix is array (Integer range <>, Integer range <>) of Scalar;

   type CompArray is array (Integer range <>) of Comp;
   type CompMatrix is array (Integer range <>, Integer range <>) of Comp;


   --SOME LINEAR ALGEBRA PROCEDURES
   --REPS
   function Scal(R: RepArray) return ScalArray;--converts a rep array into a scalar array
   function "*"(A1,A2: in RepMatrix) return RepMatrix;
   function "*"(A: RepMatrix; B: RepArray) return RepArray;
   function "*"(B: RepArray; A: RepMatrix) return RepArray;
   function "-"(B1,B2: RepArray) return RepArray;
   function "+"(B1,B2: RepArray) return RepArray;
   function "-"(A1,A2: in RepMatrix) return RepMatrix;
   function Prod(B0,B1,B2: RepArray) return Rep;
   function "*"(B1,B2: RepArray) return RepArray;
   function "*"(B1,B2: RepArray) return Rep;
   function "*"(R: Rep; A: RepMatrix) return RepMatrix;
   function "*"(R: Rep; B: RepArray) return RepArray;
   function Transpose(A: RepMatrix) return RepMatrix;
   procedure Show(B: in RepArray);

   procedure SetZero(B: out RepArray);
   procedure SetColumn(J: in Integer; V: in RepArray; A: in out RepMatrix);
   procedure SetZero(A: out RepMatrix);
   function NormInf(B: RepArray) return Rep; --returns an upper bound on \|B\|_infty
   function Norm1(B: RepArray) return Rep;
   function Norm1(A: RepMatrix) return Rep;
   function Norm2(B: RepArray) return Rep;

   --SCALARS
   procedure SetZero(B: out ScalArray);

   --COMPS
   function "-"(B: CompArray) return CompArray;
   function "+"(B1,B2: CompArray) return CompArray;
   function "*"(C: Comp;B: CompArray) return CompArray;
   function "+"(A1,A2: in CompMatrix) return CompMatrix;
   function "-"(B1,B2: CompArray) return CompArray;
   function "-"(A1,A2: in CompMatrix) return CompMatrix;
   procedure SetZero(B: out CompArray);
   procedure SetZero(A: out CompMatrix);

   procedure  PointsProdAdd(M: in CompMatrix; H: in out CompMatrix);--H=M*(H+1) pointwise
   procedure  ProdAdd(Pex: in Rep; Ro: in ScalArray; M: in CompMatrix; H: in out CompMatrix; PLpError: out Scalar);-- H=M*(H+1), PLpError=||trunc[M*(H+1)]-M*(H+1)||^p_p
   procedure  ProdAdd(M: in CompMatrix; H: in out CompMatrix);--H=M*(H+1) as Fourier series
   function NormInf(B: CompArray) return Rep; --returns an upper bound on \|B\|_infty

   function RoughLpNorm(Pex: Rep; Ro: ScalArray; B: CompMatrix) return Scalar;--returns a bound on \|B\|_p
   function UpperBoundLpNorm(Pex: Rep; Ro: ScalArray; B: CompMatrix) return Scalar; --returns an upper bound on \|B \|_p
   function EssSup(B: CompMatrix) return Scalar;--returns a bound on the esssup(B)
   procedure EssSup(B: in CompMatrix; E: out ScalArray);--E is the array of bounds of the ess sups of rows of B
   function RoughEssSup(B: CompMatrix) return Scalar;--essup(B) <= max_i sum_j |B_(i,j)|



   Scal_Pi: constant Scalar:= Scal(Rep(3.14159_26535_89793_23),Rep(3.14159_26535_89793_24));
   Scal_Zero:  constant Scalar:=Scal(Zero);
   Scal_Two:  constant Scalar:=Scal(Two);
   Co_Zero: constant Comp:=Complecs(Zero,Zero);
   Co_One: constant Comp:=Complecs(One,Zero);
   Co_I: constant Comp:=Complecs(Zero,One);
   EX: Integer:=0;
private
   Trunc: constant Boolean := (IsNumeric(Scal_Zero) and IsCoNumber(Co_Zero));

end Lin;
