with Ada.Text_IO, Reps,Reps.Ops, Reps.IO;
use Ada.Text_IO, Reps,Reps.Ops, Reps.IO;
with Round_Up;
with Ada.Strings.Unbounded;

pragma Elaborate_All (Ada.Text_IO);

generic

package RG.Beltrami is


   --SOME FFT ROUTINES
   procedure Glassman (A, B, C:   in  Integer; Data_Vector:   in out CompArray ;Inverse_Transform  :   in  Boolean:=False);--Glassman algorithm
   procedure FFT (FFT_Data:  in out CompArray; Inverse_Transform: in   Boolean:=False   );--FFT for a row
   procedure FT(H: in out CompMatrix; Inverse_Transform: in   Boolean:=False );--FFT for an array
   procedure FT_Mod(H: in out CompMatrix; Inverse_Transform: in   Boolean:=False );--FFT for the first machine


   procedure GetRow(Num_Of_Machines: in Integer;
                    ThickRad: in ScalArray;
                    Input_File: in String;
                    C: in Comp;
                    Row: out CompArray;
                    Decimal: in Boolean := True;
                    ShowErr: in  Boolean:=True);--gets a row for C from the solution of the Beltrami equation saved in Input_File.
   procedure GetRow(Num_Of_Machines: in Integer;
                    ThickRad: in ScalArray;
                    Input_File: in String;
                    R: in Scalar;
                    Row: out CompArray;
                    Decimal: in Boolean := True;
                    ShowErr: in  Boolean:=True);--gets a row for C from the solution of the Beltrami equation saved in Input_File.
   procedure GetRow(Num_Of_Machines: in Integer;
                    ThickRad: in ScalArray;
                    Input_File: in String;
                    R: in Scalar;
                    Row1,Row2: out CompArray;
                    T: out Scalar;
                    Decimal: in Boolean := True;
                    ShowErr: in  Boolean:=True);--gets a row for C from the solution of the Beltrami equation saved in Input_File.
   function G_Value(Pex: in Rep;
                    R,AEP,C: in Scalar;
                    V: in CompArray;
                    Val_At_Zero, Z: in Comp) return Comp; --g(Z),  V is the row from the solution of the Beltrami equation obtained in GetRow
   function G_Value(Pex: in Rep; R,AEP,C: in Scalar; V: in CompArray; Val_At_Zero,Z: in Comp; Rad,Psi: in Scalar) return Comp;
   function G_Value(Pex: in Rep; R,AEP,C: in Scalar; V1,V2: in CompArray; Val_At_Zero,Z: in Comp; T,Rad,Psi: in Scalar) return Comp;
   function ExpInv(C: in Comp) return Comp;--g(Rad*exp[I*Psi]


   --ROUTINES RELATED TO THE BELTRAMI EQUATUION
   procedure Simple_BeltramiSup(M: Integer; R1,R2,Eps: in Rep; Q: in Taylor; EsSup: out Scalar);
   procedure BeltramiSup(M: in Integer; ActR,R,Epsi: in Rep; Appro: in Comp; Q: in Taylor; EsSup: out Scalar);
   procedure BeltramiF(NumErr: in Rep; ThickRad: in ScalArray; Tay: in Taylor;Coeffs: out CompMatrix; FFT: in Boolean);--forms the Beltrami differential, used on the first machine only
   procedure Beltrami(NumErr: in Rep; ThickRad: in ScalArray; Tay: in Taylor;Coeffs: out CompMatrix; FFT: in Boolean);--forms the Beltrami differential, used on all machines>1
   procedure Simple_ProdAddBeltramiF(NumErr,Pex: in Rep; Ro,ThickRad: in ScalArray; Q: in Taylor; H: in out CompMatrix;Delta1,Delta2: out ScalArray; EsSup,MultError: out Scalar);
   procedure Simple_ProdAddBeltrami(NumErr,Pex: in Rep; Ro,ThickRad: in ScalArray; Q: in Taylor; H: in out CompMatrix; Delta1,Delta2: out ScalArray; EsSup,MultError: out Scalar);
   procedure ProdAddBeltramiF(NumErr,Pex: in Rep; Ro,ThickRad: in ScalArray; Q: in Taylor; H: in out CompMatrix;Delta1,Delta2: out ScalArray; EsSup,MultError: out Scalar);
   procedure ProdAddBeltrami(NumErr,Pex: in Rep; Ro,ThickRad: in ScalArray; Q: in Taylor; H: in out CompMatrix; Delta1,Delta2: out ScalArray; EsSup,MultError: out Scalar);
   procedure ZeroBelt(R0: in  Radius; Ga: in Scalar; Tay: in Taylor; ZeroValue: out Comp; VarRho: out Scalar);
   procedure PhiInv(P: in Comp; Z: in out Comp; D, DD: out Comp; Tay: in Taylor; K: in Integer);
   procedure InfBelt(R: in Scalar; P: in Comp; Tay: in Taylor; InfValue: out Comp);


   --HILBERT AND CAUCHY TRANSFORMS
   --Hilbert transforms for a linear data
   procedure Lin_T_Transform1(Iter: in Integer; R: in ScalArray; V: in  CompArray; H: in out CompMatrix; Out_Dir,CM,PM,NM: in String;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);--first machine
   procedure Lin_T_Transform2(Iter: in Integer; R: in ScalArray; V1,V2: in  CompArray; H: in out CompMatrix; Out_Dir,CM,PM,NM: in String;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);--other than first or last
   procedure Lin_T_Transform3(Iter: in Integer; R: in ScalArray; V: in  CompArray; H: in out CompMatrix;  Out_Dir,CM,PM,NM: in String;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);--last machine

   --Hilbert transforms for a constant data
   procedure Const_T_Transform1(Iter: in Integer; R,Ro,ThickR: in ScalArray; V: in  CompArray; H: in out CompMatrix; Out_Dir,CM,PM,NM: in String;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);
   procedure Const_T_Transform2(Iter: in Integer; R,Ro,ThickR: in ScalArray; V1,V2: in  CompArray; H: in out CompMatrix; Out_Dir,CM,PM,NM: in String;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);
   procedure Const_T_Transform3(Iter: in Integer; R,Ro,ThickR: in ScalArray; V: in  CompArray; H: in out CompMatrix;  Out_Dir,CM,PM,NM: in String;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);

   --Cauchy transforms for a linear data
   procedure Lin_P_Transform1(Iter: in Integer; Rad: in ScalArray; V: in  CompArray; H: in out CompMatrix;  Out_Dir,CM,PM,NM: in String;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);
   procedure Lin_P_Transform2(Iter: in Integer; Rad: in ScalArray; V1,V2: in  CompArray; H: in out CompMatrix;  Out_Dir,CM,PM,NM: in String;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);
   procedure Lin_P_Transform3(Iter: in Integer; Rad: in ScalArray; V: in  CompArray; H: in out CompMatrix;  Out_Dir,CM,PM,NM: in String;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);

   --Cauchy transforms for a constant data
   procedure Const_P_Transform1(Iter: in Integer; R,Ro,ThickR: in ScalArray; V: in  CompArray; H: in out CompMatrix; Out_Dir,CM,PM,NM: in String;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);
   procedure Const_P_Transform2(Iter: in Integer; R,Ro,ThickR: in ScalArray; V1,V2: in  CompArray; H: in out CompMatrix; Out_Dir,CM,PM,NM: in String;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);
   procedure Const_P_Transform3(Iter: in Integer; R,Ro,ThickR: in ScalArray; V: in  CompArray; H: in out CompMatrix;  Out_Dir,CM,PM,NM: in String;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);

   procedure  Iter_Fixed_Point(Pex: in Rep; N1min,Num_Of_Machines,Current_Machine,Cut_Output: in Integer;--h_(i+1)=T[mu*(h_i+1)], h_0=0
                                            F: in out CompMatrix;
                                            Rad: in ScalArray;
                                            ThickRad: in ScalArray;
                                            Ro: in ScalArray;
                                            Center: in Comp;
                                            Out_Dir: in String;
                                            Text_Output: in Boolean;
                                            Decimal: in Boolean := True;
                                            ShowErr: in  Boolean:=True);
   procedure  Iter_Cauchy(Pex: in Rep;N1min,Num_Of_Machines,Current_Machine,Cut_Output: in Integer; --F=P[mu*(h+1)]+id
                                      F: out CompMatrix;
              --                        FouErr: out Scalar;
                                      Rad: in ScalArray;
                                      ThickRad: in ScalArray;
                                      Ro: in ScalArray;
                                      Out_Dir: in String;
                                      Text_Output: in Boolean;
                                      Decimal: in Boolean := True;
                                      ShowErr: in  Boolean:=True);
   procedure  Iter_Approx_Sol(N1min,Num_Of_Machines,Current_Machine,Cut_Output: in Integer;
                              Iter: in out Integer;
                              Belt: in out CompMatrix;
                              F:  out CompMatrix;
                              Rad: in ScalArray;
                              ThickRad: in ScalArray;
                              Ro: in ScalArray;
                              Center: in Comp;
                              Con_File,Out_Dir,G_Dir,H_Dir,Mult_Dir: in String;
                              Read_Approx: in Boolean;
                              Write_H: in Boolean;
                              Decimal: in Boolean := True;
                              ShowErr: in  Boolean:=True;
                              FFT: in  Boolean:=True);

   --REGULAR READ AND WRITE PROCEDURES
   procedure   Read_Integer(F_Name: in String; I: out Integer);
   procedure   Read_Int(F_Name: in String; I: out Integer);
   procedure   Read_Vector(F_Name: in String; V: out CompArray;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);
   procedure   Read_Vec(F_Name: in String; V: out CompArray;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);


   --PARALLEL READ AND WRITE PROCEDURES
   procedure   Read(I: in Integer; Read_Log,Write_Log,Vec_Read: in String; V: in out CompArray;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);
   procedure   Read_B(Write_Log,Comp_Read: in String; V: out Comp;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);
   procedure   Read_T(I: in Integer; Read_Log,Write_Log,Vec_Read: in String; V: in out CompArray;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);
   procedure   Read_P(I: in Integer; Read_Log,Write_Log,Vec_Read: in String; V: in out CompArray;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);
   procedure  Read_SB(H: out CompMatrix; File_Name: in String; In_Values,Out_Values: in Boolean;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);
   procedure  Read_Part(H: out CompMatrix; File_Name: in String; In_Values,Out_Values: in Boolean;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);
   procedure  Read_Renorm(H: out CompMatrix; N: in Integer;  File_Name: in String; In_Values,Out_Values: in Boolean;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);
   procedure  Write(I: in Integer; Read_Log,Write_Log,Vec_Write: in String; V: in CompArray;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);
   procedure  Write_B(Write_Log, Comp_Write: in String; V: in Comp;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);
   procedure  Write_T(I: in Integer; Read_Log,Write_Log, Vec_Write: in String; V: in CompArray;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);
   procedure  Write_SB(H: in CompMatrix; File_Name: in String;  Decimal: in Boolean := True; ShowErr: in  Boolean:=True);
   function Read_N1(File_Name: in String) return Integer;
   function Read_N2(File_Name: in String) return Integer;
   function Read_M(File_Name: in String) return Integer;
   function Set_Machine(This_Machine,Num_Of_Machines: Integer) return String;


end RG.Beltrami;
