with Messages,  Ada.Text_IO,Ada.Integer_Text_IO,Ada.IO_Exceptions,Intervals.IO, Ada.Real_Time, Ada.Calendar, Ada.Strings.Unbounded;
use Messages, Ada.Text_IO,Ada.Integer_Text_IO, Ada.Strings.Unbounded;

pragma Elaborate_All (Messages);
package body RG is

   pragma Suppress(Storage_Check);

   function IMin(I1,I2: Integer) return Integer renames Integer'Min;
   function IMax(I1,I2: Integer) return Integer renames Integer'Max;
   pragma Inline (Imin,Imax);
   pragma Suppress(Storage_Check);



   --ROUTINES RELATED TO THE CONSTRUCTION OF THE FUNDAMENTAL CRESCENT
   function Val(Lambda: Parametrization) return Comp is
   begin
     return Lambda.C;
   end Val;


   function TauInv(C: Comp) return Comp is
     Z1,Z2,F,DZ1,DZ2,DZ3,DZ4: Comp;
   begin
      Z1:=(C-P);
      Z2:=(Cn2-P);
      F:=Neg(CoI)*( Log( ResetErr(Z1*Cn2) )-Log( ResetErr(C*Z2) ) ) /LogA;


      if not Trunc then
         DZ1:=Inv(Z1);
         DZ2:=Inv(Z2);
         DZ3:=Inv(C);
         DZ4:=Inv(Cn2);
         F:=F+Ball0(ErrPart(Z1)*DZ1/LogA)+
           Ball0(ErrPart(Z2)*DZ2/LogA)+
           Ball0(ErrPart(C)*DZ3/LogA)+
           Ball0(ErrPart(Cn2)*DZ4/LogA);
      end if;
      return F;
   end TauInv;


   function Tau_Log(C: Comp) return Comp is
      F: Comp;
   begin
      F:=Log( ResetErr(C) );

      if not Trunc then
         F:=F+Ball0(ErrPart(C)*Inv(C));
      end if;
      return F;
   end Tau_Log;


   function Tau(C: Comp) return Comp is
   begin
      return  P/(CoOne-Exp(CoI*(LogA*C+LogBB)));
   end Tau;


   procedure Parametrize_1(R: in Scalar; C: out Comp; Der,RDer: out Comp) is --FINE
      R1,Der_T,RDer_T,S,T,A,B,G,D,E,SL1,SL2,Xc,Yc,Xp,Yp: Scalar;
      N: constant Integer:=3;
   begin
      --These are the tangents of the lines along which two parabolas, constituting a boundary of the fundamental crescent, meet. Adjust accordingly.
      SL1:=Scal(-1.1);
      SL2:=Scal(-1.1);

      Xc:=RePart(Cn2n);
      Xp:=RePart(P);
      Yc:=ImPart(Cn2n);
      Yp:=ImPart(P);

      R1:=Sqrt(R); --depends on R
      E:=AbsVal(Cn2n)+AbsVal(Cn2n-P);
      T:=E*R1/(R1+ScalOne); --depends on R
      Der_T:=Inv(R1)*E/(Two*(R1+ScalOne)*(R1+ScalOne)); --depends on R
      RDer_T:=R1*E/(Two*(R1+ScalOne)*(R1+ScalOne)); -- R*Der_T,  depends on R

      S:=AbsVal(Cn2n);
      if Approx(S)<Zero then
         New_Line;
         Message("RG_Ops.Parametrize",Parameter_Error);
      end if;
      if Approx(T)<Approx(S) then
         --Parametrization of the "bottom" with a parabola
         A:=(Xc*SL1-Yc)/Sqr(Xc);
         B:=-(Xc*SL1-Two*Yc)/Xc;
         G:=T*Xc/S;
         C:=SetBall(G,A*Sqr(G)+B*G);
         Der:=Xc/S*Der_T*SetBall(ScalOne,Two*A*G+B);
         RDer:=Xc/S*RDer_T*SetBall(ScalOne,Two*A*G+B);
      else
         --Parametrization of the "top" with a parabola
         A:=-(Yp+Xc*SL2-SL2*Xp-Yc)/(SL2*Sqr(Yp-Yc));
         B:=(Two*Yc*Xc*SL2-Two*Yc*Xp*SL2-Sqr(Yc)+Sqr(Yp))/(SL2*Sqr(Yp-Yc));
         D:=(Sqr(Yc)*Yp+Sqr(Yc)*Xp*SL2-Yc*Sqr(Yp)+Xc*Sqr(Yp)*SL2-Two*Xc*Yp*Yc*SL2)/(SL2*Sqr(Yp-Yc));
         G:=Yc*(E-T)/AbsVal(Cn2n-P)+Yp*(T-S)/AbsVal(Cn2n-P);
         C:=SetBall(A*Sqr(G)+B*G+D,G);
         Der:=(Yp-Yc)/AbsVal(Cn2n-P)*Der_T*SetBall(Two*A*G+B,ScalOne);
         RDer:=(Yp-Yc)/AbsVal(Cn2n-P)*RDer_T*SetBall(Two*A*G+B,ScalOne);
      end if;
   end Parametrize_1;


   procedure Parametrize_2(R: in Scalar; C,Der,RDer,RDerDer,CC,DDer,RRDer,RRDerDer: out Comp) is --FINE
      R1,RR1,Der_T,Der_TT,RDer_T,RDer_TT,RDerDer_T,RDerDer_TT,S,T,A,B,G,GG,TT,D,E,SL1,SL2,Xc,Yc,Xp,Yp: Scalar;
   begin
      --These are the tangents of the lines along which two parabolas, constituting a boundary of the fundamental crescent, meet. Adjust accordingly.
         --Qn=3
      SL1:=Scal(-1.1);
      SL2:=Scal(-1.1);
      Xc:=RePart(Cn2n);
      Xp:=RePart(P);
      Yc:=ImPart(Cn2n);
      Yp:=ImPart(P);
      E:=AbsVal(Cn2n)+AbsVal(Cn2n-P);

      R1:=Sqrt(Scal(Approx(R),Approx(R)+Err(R))); --depends on R
      T:=E*R1/(R1+ScalOne); --depends on R
      Der_T:=Inv(R1)*E/(Two*(R1+ScalOne)*(R1+ScalOne)); --depends on R
      RDer_T:=R1*E/(Two*(R1+ScalOne)*(R1+ScalOne)); -- R*Der_T,  depends on R
      RDerDer_T:=(-E)*(ScalOne+Three*R1)/(Four*R1*Sqr(R1+ScalOne)*(R1+ScalOne));

      RR1:=Sqrt(Center(R)); --depends on R
      TT:=E*RR1/(RR1+ScalOne); --depends on R
      Der_TT:=Inv(RR1)*E/(Two*(RR1+ScalOne)*(RR1+ScalOne)); --depends on R
      RDer_TT:=RR1*E/(Two*(RR1+ScalOne)*(RR1+ScalOne)); -- R*Der_T,  depends on R
      RDerDer_TT:=(-E)*(ScalOne+Three*RR1)/(Four*RR1*Sqr(RR1+ScalOne)*(RR1+ScalOne));

      S:=AbsVal(Cn2n);
      if Inf(S)<Zero then
         New_Line;
         Message("RG_Ops.Parametrize",Parameter_Error);
      end if;
      if Approx(T)<Approx(S) then
         --Parametrization of the "bottom" with a parabola
         A:=(Xc*SL1-Yc)/Sqr(Xc);
         B:=-(Xc*SL1-Two*Yc)/Xc;
         G:=T*Xc/S;
         GG:=TT*Xc/S;
         C:=SetBall(GG,A*Sqr(GG)+B*GG);

         Der:=Xc/S*Der_TT*SetBall(ScalOne,Two*A*GG+B);
         RDer:=Xc/S*RDer_TT*SetBall(ScalOne,Two*A*GG+B);
         RDerDer:=Xc/S*(RDerDer_TT*SetBall(ScalOne,Two*A*GG+B)+Xc/S*RDer_TT*Der_TT*SetBall(ScalZero,Two*A));

         DDer:=Xc/S*Der_T*SetBall(ScalOne,Two*A*G+B);
         RRDer:=Xc/S*RDer_T*SetBall(ScalOne,Two*A*G+B);
         RRDerDer:=Xc/S*(RDerDer_T*SetBall(ScalOne,Two*A*G+B)+Xc/S*RDer_T*Der_T*SetBall(ScalZero,Two*A));
      else
         --Parametrization of the "top" with a parabola
         A:=-(Yp+Xc*SL2-SL2*Xp-Yc)/(SL2*Sqr(Yp-Yc));
         B:=(Two*Yc*Xc*SL2-Two*Yc*Xp*SL2-Sqr(Yc)+Sqr(Yp))/(SL2*Sqr(Yp-Yc));
         D:=(Sqr(Yc)*Yp+Sqr(Yc)*Xp*SL2-Yc*Sqr(Yp)+Xc*Sqr(Yp)*SL2-Two*Xc*Yp*Yc*SL2)/(SL2*Sqr(Yp-Yc));
         G:=Yc*(E-T)/AbsVal(Cn2n-P)+Yp*(T-S)/AbsVal(Cn2n-P);
         GG:=Yc*(E-TT)/AbsVal(Cn2n-P)+Yp*(TT-S)/AbsVal(Cn2n-P);
         C:=SetBall(A*Sqr(GG)+B*GG+D,GG);

         Der:=(Yp-Yc)/AbsVal(Cn2n-P)*Der_TT*SetBall(Two*A*GG+B,ScalOne);
         RDer:=(Yp-Yc)/AbsVal(Cn2n-P)*RDer_TT*SetBall(Two*A*GG+B,ScalOne);
         RDerDer:=(Yp-Yc)/AbsVal(Cn2n-P)*(RDerDer_TT*SetBall(Two*A*GG+B,ScalOne)+
                                          (Yp-Yc)/AbsVal(Cn2n-P)*RDer_TT*Der_TT*SetBall(Two*A,ScalZero));

         DDer:=(Yp-Yc)/AbsVal(Cn2n-P)*Der_T*SetBall(Two*A*G+B,ScalOne);
         RRDer:=(Yp-Yc)/AbsVal(Cn2n-P)*RDer_T*SetBall(Two*A*G+B,ScalOne);
         RRDerDer:=(Yp-Yc)/AbsVal(Cn2n-P)*(RDerDer_T*SetBall(Two*A*G+B,ScalOne)+
                                           (Yp-Yc)/AbsVal(Cn2n-P)*RDer_T*Der_T*SetBall(Two*A,ScalZero));
      end if;
      CC:=C+Scal(-Err(R),Err(R))*DDer;
   end Parametrize_2;


   procedure SInvH1(FCAppr, Appr: in out Comp; C: in Comp; Tay: in Taylor; Er: in Radius) is
      DerTau,DerTauInv,DerF,C1PC1,C2PC2,DerIF,DerG,C1,C2,H1_y,H1_x,RDerG: Comp;
      C3: Comp:=FCAppr;
      M,I: Integer;
      IDh,A: ScalMatrix(1..2,1..2);
      V,X,Xold: ScalArray(1..2);
      NormA,Det,Y,XX: Scalar;
      Diff,X1,Ro,Eps: Rep;
      InvLogA: constant Scalar:=Inv(LogA);
   begin
      X(1):=RePart(Appr);
      X(2):=ImPart(Appr);
      Diff:=Two*Er;
      --Computing x_0 and x_1, see Kantorovich theorem
      I:=0;
      while Diff > Er loop
         I:=I+1;
         X(1):=Center(X(1));
         X(2):=Center(X(2));
         X1:=Approx(X(1));
         XX:=Scal( Modulo( Approx(X(1)) , One ) );
         Y:=Exp(Two*ScalPi*X(2));
         Parametrize_1(Y,C2,DerG,RDerG);
         C1:=C2;
         Newton(Qn,Rho,C1,Tay,C2);
         DerG:=Two*ScalPi*Y*DerG;
         Ro:=SupRad(Qn,Rho,C1,Tay);
         DerIF:=Inv(DerVal(Qn,Rho,Ro,C1,Tay));
         C2PC2:=(C2-P)*C2; C1PC1:=(C1-P)*C1;

         --derivatives
         H1_x:=InvLogA*Neg(CoI)*Log( (C2-P)*C1/((C1-P)*C2) ); --h_x
         H1_y:=InvLogA*DerG*P*Neg(CoI)*((ScalOne-XX)*Inv(C1PC1)*DerIF+XX*Inv(C2PC2)); --h_y
         if X1<Zero then
            M:=0;
            while X1<Zero loop
               X1:=X1+One;
               M:=M+1;
            end loop;
            C2:=SH1(FCAppr,XX,X(2),Tay);
            DerTau:=LogA*CoI*P*Exp(CoI*(LogA*C2+LogBB))/Sqr(CoOne-Exp(CoI*(LogA*C2+LogBB)));
            C2:=Tau(C2);
            C3:=FCAppr;
            Newton(M*Qn,Rho,C3,Tay,C2);
            FCAppr:=C3;
            Ro:=SupRad(M*Qn,Rho,C3,Tay);
            DerF:=DerVal(M*Qn,Rho,Ro,C3,Tay);
            DerTauInv:=Neg(CoI)*P/(LogA*C3*(C3-P));
            H1_x:=DerTau*DerTauInv*H1_x/DerF;
            H1_y:=DerTau*DerTauInv*H1_y/DerF;
         elsif X1>=One then
            M:=0;
            while X1>=One loop
               X1:=X1-One;
               M:=M+1;
            end loop;
            C2:=SH1(FCAppr,XX,X(2),Tay);
            DerTau:=LogA*CoI*P*Exp(CoI*(LogA*C2+LogBB))/Sqr(CoOne-Exp(CoI*(LogA*C2+LogBB)));
            C2:=Tau(C2);
            Ro:=SupRad(M*Qn,Rho,C2,Tay);
            DerF:=DerVal(M*Qn,Rho,Ro,C2,Tay);
            C2:=Val(M*Qn,Rho,C2,Tay);
            FCAppr:=C2;
            DerTauInv:=Neg(CoI)*P/(LogA*C2*(C2-P));
            H1_x:=DerTauInv*DerF*DerTau*H1_x;
            H1_y:=DerTauInv*DerF*DerTau*H1_y;
         else
            null;
         end if;

         IDh(2,2):=RePart(H1_x); -- u_x
         IDh(2,1):=-ImPart(H1_x); -- -v_x
         IDh(1,2):=-RePart(H1_y);-- -u_y
         IDh(1,1):=ImPart(H1_y);--v_y
         Det:=IDh(1,1)*IDh(2,2)-IDh(2,1)*IDh(1,2);
         C1:=C-SH1(FCAppr,X(1),X(2),Tay);
         Xold:=X;
         V(1):=RePart(C1);
         V(2):=ImPart(C1);
         X(1):=Scal(Approx(X(1)+(IDh(1,1)*V(1)+IDh(1,2)*V(2))/Det));
         X(2):=Scal(Approx(X(2)+(IDh(2,1)*V(1)+IDh(2,2)*V(2))/Det));
         Diff:=Max(Sup(abs( X(1)-Xold(1) )),Sup(abs(X(2)-Xold(2) )) );
         if I>100 then
            New_Line;
            Message("RG.SInvH1",Giving_Up);
         end if;
      end loop;
      X(1):=Xold(1)+(IDh(1,1)*V(1)+IDh(1,2)*V(2))/Det;
      X(2):=Xold(2)+(IDh(2,1)*V(1)+IDh(2,2)*V(2))/Det;
      Appr:=SetBall(X(1),X(2));



      --error estimates: define the newton map
      -- N(h)=h+g_1(x_0+M h)-C, M=(-Dg_1)^{-1}
      -- show that it contracts a disk around x_0

      if Trunc then return; end if;

      Eps:=Rep(10.0E-20);

      Diff:=One;
      while Diff>=Zero loop

         X(1):=Center(X(1))+Scal(-Eps,Eps);
         X(2):=Center(X(2))+Scal(-Eps,Eps);

         X1:=Approx(X(1));
         XX:=Scal( Modulo( Approx(X(1)) , One ) )+Scal(-Eps,Eps);
         Y:=Exp(Two*ScalPi*X(2));
         Parametrize_1(Y,C2,DerG,RDerG);
         C1:=C2;
         Newton(Qn,Rho,C1,Tay,C2);
         DerG:=Two*ScalPi*Y*DerG;
         Ro:=SupRad(Qn,Rho,C1,Tay);
         DerIF:=Inv(DerVal(Qn,Rho,Ro,C1,Tay));
         C2PC2:=(C2-P)*C2; C1PC1:=(C1-P)*C1;

         H1_x:=InvLogA*Neg(CoI)*Log( (C2-P)*C1/((C1-P)*C2) ); --h_x
         H1_y:=InvLogA*DerG*P*Neg(CoI)*((ScalOne-XX)*Inv(C1PC1)*DerIF+XX*Inv(C2PC2)); --h_y
         if X1<Zero then
            M:=0;
            while X1<Zero loop
               X1:=X1+One;
               M:=M+1;
            end loop;
            C2:=SH1(FCAppr,XX,X(2),Tay);
            DerTau:=LogA*CoI*P*Exp(CoI*(LogA*C2+LogBB))/Sqr(CoOne-Exp(CoI*(LogA*C2+LogBB)));
            C2:=Tau(C2);
            C3:=FCAppr;
            Newton(M*Qn,Rho,C3,Tay,C2);
            FCAppr:=C3;
            Ro:=SupRad(M*Qn,Rho,C3,Tay);
            DerF:=DerVal(M*Qn,Rho,Ro,C3,Tay);
            DerTauInv:=Neg(CoI)*P/(LogA*C3*(C3-P));
            H1_x:=DerTau*DerTauInv*H1_x/DerF;
            H1_y:=DerTau*DerTauInv*H1_y/DerF;
         elsif X1>=One then
            M:=0;
            while X1>=One loop
               X1:=X1-One;
               M:=M+1;
            end loop;
            C2:=SH1(FCAppr,XX,X(2),Tay);
            DerTau:=LogA*CoI*P*Exp(CoI*(LogA*C2+LogBB))/Sqr(CoOne-Exp(CoI*(LogA*C2+LogBB)));
            C2:=Tau(C2);
            Ro:=SupRad(M*Qn,Rho,C2,Tay);
            DerF:=DerVal(M*Qn,Rho,Ro,C2,Tay);
            C2:=Val(M*Qn,Rho,C2,Tay);
            FCAppr:=C2;
            DerTauInv:=Neg(CoI)*P/(LogA*C2*(C2-P));
            H1_x:=DerTauInv*DerF*DerTau*H1_x;
            H1_y:=DerTauInv*DerF*DerTau*H1_y;
         else
            null;
         end if;

         --A bound on DN
         A(1,1):=ScalOne-(RePart(H1_x)*IDh(1,1)+RePart(H1_Y)*IDh(2,1))/Det;
         A(1,2):=-(RePart(H1_x)*IDh(1,2)+RePart(H1_Y)*IDh(2,2))/Det;
         A(2,1):=-(ImPart(H1_x)*IDh(1,1)+ImPart(H1_Y)*IDh(2,1))/Det;
         A(2,2):=ScalOne-(ImPart(H1_x)*IDh(1,2)+ImPart(H1_Y)*IDh(2,2))/Det;

         NormA:=Scal(Max(SupAbs(Half*(A(1,1)+A(2,2)+Exp(Half*Log(Rep(4.0)*A(1,2)*A(2,1)+Sqr(A(1,1)-A(2,2)))))),SupAbs(Half*(A(1,1)+A(2,2)-Exp(Half*Log(Rep(4.0)*A(1,2)*A(2,1)+Sqr(A(1,1)-A(2,2)))))) ));

         --A bound on N(0)
         C1:=C-SH1(FCAppr,Center(X(1)),Center(X(2)),Tay);

         --Verification of contraction
         Diff:=Sup(AbsVal(C1)-Eps*(ScalOne-NormA));
         Appr:=SetBall(Center(X(1)),Center(X(2)))+SetBall(CoZero,Eps);
         Eps:=Two*Eps;
      end loop;
   end SInvH1;


   procedure SInvH1(FCAppr, Appr: in out Comp; C: in Comp; IDh: in ScalMatrix; Tay: in Taylor; Er: in Radius) is
      C1: Comp;
      V,X,Xold: ScalArray(1..2);
      Det: Scalar;
      Diff: Rep;
      InvLogA: constant Scalar:=Inv(LogA);
   begin
      X(1):=RePart(Appr);
      X(2):=ImPart(Appr);
      Diff:=Two*Er;
      --Computing x_0 and x_1, see Kantorovich theorem
      while Diff > Er loop
         X(1):=Center(X(1));
         X(2):=Center(X(2));
         Det:=IDh(1,1)*IDh(2,2)-IDh(2,1)*IDh(1,2);
         C1:=C-SH1(FCAppr,X(1),X(2),Tay);
         Xold:=X;
         V(1):=RePart(C1);
         V(2):=ImPart(C1);
         X(1):=Scal(Approx(X(1)+(IDh(1,1)*V(1)+IDh(1,2)*V(2))/Det));
         X(2):=Scal(Approx(X(2)+(IDh(2,1)*V(1)+IDh(2,2)*V(2))/Det));
         Diff:=Max(Sup(abs( X(1)-Xold(1) )),Sup(abs(X(2)-Xold(2) )) );
      end loop;
      X(1):=Xold(1)+(IDh(1,1)*V(1)+IDh(1,2)*V(2))/Det;
      X(2):=Xold(2)+(IDh(2,1)*V(1)+IDh(2,2)*V(2))/Det;
      Appr:=SetBall(X(1),X(2));
   end SInvH1;


   procedure DerSH1(C: in Comp; IDh: out ScalMatrix; Tay: in Taylor) is
      C1PC1,C2PC2,DerIF,DerG,C1,C2,H1_y,H1_x,RDerG: Comp;
      X: ScalArray(1..2);
      Y: Scalar;
      InvLogA: constant Scalar:=Inv(LogA);
   begin
      X(1):=RePart(C);
      X(2):=ImPart(C);
      Y:=Exp(Two*ScalPi*X(2));
      Parametrize_1(Y,C1,DerG,RDerG);
      C2:=Val(Qn,Rho,C1,Tay);
      DerG:=Two*ScalPi*Y*DerG;

      DerIF:=DerVal(Qn,Rho,SupMod(C1),C1,Tay);
      C2PC2:=(C2-P)*C2; C1PC1:=(C1-P)*C1;

      --derivatives
      H1_x:=InvLogA*Neg(CoI)*Log( (C2-P)*C1/((C1-P)*C2) ); --h_x
      H1_y:=InvLogA*DerG*P*Neg(CoI)*((ScalOne-X(1))*Inv(C1PC1)+X(1)*Inv(C2PC2)*DerIF); --h_y
      IDh(2,2):=RePart(H1_x); -- u_x
      IDh(2,1):=-ImPart(H1_x); -- -v_x
      IDh(1,2):=-RePart(H1_y);-- -u_y
      IDh(1,1):=ImPart(H1_y);--v_y
   end DerSH1;


   function SH1(Appro: Comp; X,Y: Scalar; Tay: Taylor) return Comp is
      PP,RDerr,Derr,C1,C2,L1,L2: Comp;
      XX: Scalar:=Scal(Modulo(Approx(X),One));
      X1: Rep:=Approx(X);
      C: Comp:=Appro;
      M: Integer;
   begin
      --Appro is an approximate C1 in Newton, not used right now
      Parametrize_1(Exp(Two*ScalPi*Y),C2,Derr,RDerr);
      C1:=C2;
      Newton(Qn,Rho,C1,Tay,C2);
      L1:=TauInv(C1);
      L2:=TauInv(C2);
      PP:=(ScalOne-XX)*L1+XX*L2;

      if X1<Zero then
         M:=0;
         while X1 <Zero loop
            X1:=X1+One;
            M:=M+1;
         end loop;
         PP:=Tau(PP);
         C:=Appro;
         Newton(M*Qn,Rho,C,Tay,PP);
         PP:=C;
         PP:=TauInv(PP);
      elsif X1>=One then
         PP:=Tau(PP);
         while X1>=One loop
            PP:=Val(Qn,Rho,PP,Tay);
            X1:=X1-One;
         end loop;
         PP:=TauInv(PP);
      else
       null;
      end if;
      return PP;
   end SH1;


   procedure  SH1(Appro: in out Comp; PP: out Comp; X,Y: in Scalar; Tay: in Taylor) is
      RDerr,Derr,C1,C2,L1,L2: Comp;
      XX: Scalar:=Scal(Modulo(Approx(X),One));
      X1: Rep:=Approx(X);
      C: Comp;
      M: Integer;
   begin
      --Appro is an approximate C1 in Newton, not used right now
      Parametrize_1(Exp(Two*ScalPi*Y),C1,Derr,RDerr);
      C2:=Val(Qn,Rho,C1,Tay);
      L1:=TauInv(C1);
      L2:=TauInv(C2);
      PP:=(ScalOne-XX)*L1+XX*L2;
      if X1<Zero then
         M:=0;
         while X1<Zero loop
            X1:=X1+One;
            M:=M+1;
        end loop;
        PP:=Tau(PP);
        C:=Appro;
        Newton(M*Qn,Rho,C,Tay,PP);
        PP:=C;
        PP:=TauInv(PP);
      elsif X1>=One then
         PP:=Tau(PP);
        while X1>One loop
           PP:=Val(Qn,Rho,PP,Tay);
           X1:=X1-One;
        end loop;
        PP:=TauInv(PP);
      end if;
   end SH1;


   procedure  ComputeTay(N: in Integer; Cont,FCont: in CompArray; Tay: out Taylor) is
      J1,J2: Integer;
      Summa,C3,C4,C5: Comp;
   begin
      SetZero(Tay);
      for I in 0..N loop
         Summa:=CoZero;
         for J in Cont'Range loop
            if J=Cont'First then
               J1:=J;J2:=Cont'Last;
            else
               J1:=J;J2:=J-1;
            end if;
            C3:=CoOne;
            C4:=CoOne;
            C5:=CoOne;
            for II in 1..I loop
               C3:=C3/(Cont(J1));
               C4:=Two*C4/(Cont(J2)+Cont(J1));
               C5:=C5/(Cont(J2));
            end loop;
            if I/=0 then
               Summa:=Summa-(FCont(J1)/Rep(I))*(C3-C4)-(FCont(J2)/Rep(I))*(C4-C5);
            end if;
         end loop;
         SetComponent(I,Inv(ScalTwoPi*CoI)*Summa,Tay);
      end loop;
   end ComputeTay;


   procedure SetCont(Cont_Length,This_Machine,Num_Of_Machines: in Integer; First,Last: out Integer) is
      Incr: Integer;
   begin
      Incr:=Cont_Length/Num_Of_Machines;
      First:=1;
      Last:=First+Incr-1;
      for I in 2..This_Machine loop
         First:=First+Incr;
         Last:=First+Incr-1;
      end loop;
      if This_Machine=Num_Of_Machines then
         Last:=Cont_Length;
      end if;
   end SetCont;

end RG;



















