with Ada.Text_IO, Reps,Reps.Ops, Reps.IO, Lin;
use Ada.Text_IO, Reps,Reps.Ops, Reps.IO;
with Taylors, Taylors.Ops, Taylors.IO;
with Round_Up;
with Ada.Strings.Unbounded;
pragma Elaborate_All (Ada.Text_IO, Taylors.Ops, Taylors.IO, Lin);

generic

   Pdeg: in Positive;
   Dho: in Natural;
   type Scalar is private;
   type Comp is private;


   --intervals
   with function IsNumeric(S: Scalar) return Boolean is <>;
   with function Scal(K: Integer) return Scalar is <>;
   with function Scal(R: Rep) return Scalar is <>;
   with function Err0(S: Scalar) return Boolean is <>;
   with function Err(S: Scalar) return Rep is <>;
   with function Approx(S: Scalar) return Rep is <>;
   with function "-"(S: Scalar) return Scalar is <>;
   with function Center(S: Scalar) return Scalar is <>;
   with procedure ResetCenter(S: in out Scalar) is <>;
   with function ResetCenter(S: Scalar) return Scalar is <>;

   with function Scal(R1,R2: Rep) return Scalar is <>;
   with function Ball(S: Scalar) return Scalar is <>;
   with function Ball(S: Rep) return Scalar is <>;
   with function Inf(S: Scalar) return Rep is <>;
   with function Sup(S: Scalar) return Rep is <>;
   with function SupAbs(S: Scalar) return Rep is <>;
   with function "<"(S1,S2: Scalar) return Boolean is <>;
   with function IntFloor(S: Scalar) return Integer is <>;
   with function IntCeiling(S: Scalar) return Integer is <>;
   with function "abs"(S: Scalar) return Scalar is <>;
   with function Min(S1,S2: Scalar) return Scalar is <>;
   with function Max(S1,S2: Scalar) return Scalar is <>;
   with function Cap(R: Radius; S: Scalar) return Scalar is <>;
   with function Up(R: Rep; Dummy: Scalar) return Rep is <>;
   with procedure ErrMult(R: in Rep; S: in out Scalar) is <>;

   with function "+"(S,T: Scalar) return Scalar is <>;
   with function "-"(S,T: Scalar) return Scalar is <>;
   with function "*"(R: Rep; S: Scalar) return Scalar is <>;
   with function "*"(S: Scalar; R: Rep) return Scalar is <>;
   with function "*"(S,T: Scalar) return Scalar is <>;
   with function "/"(S: Scalar; R: Rep) return Scalar is <>;
   with function "/"(R: Rep; S: Scalar) return Scalar is <>;
   with function "/"(S,T: Scalar) return Scalar is <>;
   with function Sqr(S: Scalar) return Scalar is <>;
   with function Inv(S: Scalar) return Scalar is <>;
   with function "**"(S: Scalar; I: Integer) return Scalar is <>;
   with function Short_Exp(R: Scalar; Iter: Integer) return Scalar is <>;

   with function Cosh(S: Scalar) return Scalar is <>;
   with function Sinh(S: Scalar) return Scalar is <>;
   with function Sin(S: Scalar) return Scalar is <>;
   with function Cos(S: Scalar) return Scalar is <>;
   with function Exp(S: Scalar) return Scalar is <>;
   with function Log(S: Scalar) return Scalar is <>;
   with function Sqrt(S: Scalar) return Scalar is <>;
   with procedure CosSin(S: in Scalar; Sc,Ss: out Scalar) is <>;

   with procedure Show(S: in Scalar; NewLine: in Boolean := True) is <>;
   with procedure Get(F: in File_Type; S: out Scalar; Decimal: in Boolean := True) is <>;
   with procedure Put(F: in File_Type; S: in Scalar; Decimal: in Boolean := True) is <>;

   --balls

   with function RePart(C: Comp) return Scalar is <>;
   with function ImPart(C: Comp) return Scalar is <>;
   with function Conj(C: Comp) return Comp is <>;
   with function SetBall(C: Comp;R: Rep:=Zero) return Comp is <>;
   with function IsCoNumber(S: Comp) return Boolean is <>;
   with function IsBall(S: Comp) return Boolean is <>;
   with function Complecs(Re,Im: Rep) return Comp is <>;
   with procedure ResetCenter(S: in out Comp) is <>;
   with function ResetCenter(S: Comp) return Comp is <>;
   with procedure Split(S: in Comp; Sr,Se: out Comp) is <>;
   with procedure ResetErr(S: in out Comp) is <>;
   with function ResetErr(S: Comp) return Comp is <>;
   with function CenterBall(S: Comp) return Comp is <>;
   with function Ball0(R: Rep) return Comp is <>;
   with function Ball0(R: Comp) return Comp is <>;
   with function ErrPart(S: Comp) return Rep is <>;

   with function InfRe(S: Comp) return Rep is <>;
   with function SupRe(S: Comp) return Rep is <>;
   with function InfIm(S: Comp) return Rep is <>;
   with function SupIm(S: Comp) return Rep is <>;
   with function SupMod(S: Comp) return Rep is <>;
   with function InfMod(S: Comp) return Rep is <>;
   with function AbsBall(S: Comp) return Comp is <>;
   with function AbsVal(S: Comp) return Scalar is <>;
   with function Arg(C: Comp) return Scalar is <>;
   with procedure ErrMult(R: in Rep; S: in out Comp) is <>;
   with function Neg(S: Comp) return Comp is <>;
   with procedure RandomBall(Seed: in out Long_Integer; S: out Comp) is <>;

   with function Contains(S: Comp; C: Comp) return Boolean is <>;
   with function Intersects(S: Comp; T: Comp) return Boolean is <>;
   with function SetBall(R1,R2: Scalar) return Comp is <>;
   with function "+"(S,T: Comp) return Comp is <>;
   with function "-"(S,T: Comp) return Comp is <>;
   with function "*"(R: Rep; S: Comp) return Comp is <>;
   with function "/"(S: Comp; R: Rep) return Comp is <>;
   with function "*"(R: Scalar; S: Comp) return Comp is <>;
   with function "/"(S: Comp; R: Scalar) return Comp is <>;
   with function Sqr(S: Comp) return Comp is <>;
   with function "*"(S,T: Comp) return Comp is <>;
   with function Inv(S: Comp) return Comp is <>;
   with function Long_Inv(S: Comp) return Comp is <>;
   with function "/"(S,T: Comp) return Comp is <>;
   with function "**"(S: Comp; I: Integer) return Comp is <>;
   with function Cap(R: Radius; S: Comp) return Comp  is <>;
   with function MaxMod(S1,S2: Comp) return Comp is <>;

   with function Cosh(S: Comp) return Comp is <>;
   with function Sinh(S: Comp) return Comp is <>;
   with function Exp(S: Comp) return Comp is <>;
   with function EiExp(S1,S2: Comp; NumErr: Rep:=Rep(1.0E-15)) return Comp is <>;
   with function EiExpp(S0,S1,S2: Comp; NumErr: Rep:=Rep(1.0E-15)) return Comp is <>;
   with function Cos(S: Comp) return Comp is <>;
   with function Sin(S: Comp) return Comp is <>;
   with function Log(S: Comp) return Comp is <>;
   with function Logg(S0,S: Comp) return Comp is <>;
   with function Sqrt(S: Comp) return Comp is <>;
   --   with function PiBall return Comp is <>;

   with procedure Show(S: in Comp; Cut: in Rep:=Zero) is <>;
   with procedure Show(Text: in String; S: in Comp; Cut: in Rep:=Zero) is <>;
   with procedure Get(F: in File_Type; S: out Comp; Decimal: in Boolean := True; ShowErr: in  Boolean:=True) is <>;
   with procedure Put(F: in File_Type; S: in Comp; Decimal: in Boolean := True; ShowErr: in  Boolean:=True) is <>;

package RG is

   package RG_Tay is new Taylors (Pdeg,Dho,Scalar,Comp);
   package RG_Tay_Ops  is new RG_Tay.Ops;
   package Rg_Tay_IO is new RG_Tay.IO;
   use RG_Tay, RG_Tay_Ops,RG_Tay_IO;

   package RG_Lin is new Lin(Scalar,Comp);
   use RG_Lin;

   type Parametrization is private;
   type TayArray is array (Integer range <>) of Taylor;
   type ParamArray is array (Integer range <>) of Parametrization;

   --ROUTINES RELATED TO THE CONSTRUCTION OF THE FUNDAMENTAL CRESCENT
   function Val(Lambda: Parametrization) return Comp;
   function Tau(C: Comp) return Comp;--tau
   function TauInv(C: Comp) return Comp;--inverse tau
   function Tau_Log(C: Comp) return Comp;--log the clever way
   procedure Parametrize_1(R: in Scalar; C:  out Comp; Der,RDer: out Comp);--Parametrization C(r) of one side of the fundamental crescent, Der=C'(r),RDer=r*C'(r)
   procedure Parametrize_2(R: in Scalar; C,Der,RDer,RDerDer,CC,DDer,RRDer,RRDerDer: out Comp);--Parametrization C(r) of one side of the fundamental crescent, Der=C'(r),RDer=r*C'(r),RDerDer=r*C''(r),RRDerDer=r^2*C''(r)
   procedure SInvH1(FCAppr, Appr: in out Comp; C: in Comp; Tay: in Taylor; Er: in Radius);--inverse g_1
   procedure SInvH1(FCAppr, Appr: in out Comp; C: in Comp; IDh: in ScalMatrix; Tay: in Taylor; Er: in Radius);--inverse g_1
   procedure DerSH1(C: in Comp; IDh: out ScalMatrix; Tay: in Taylor);--inverse g_1
   function SH1(Appro: Comp; X,Y: Scalar; Tay: Taylor) return Comp;--g_1
   procedure SH1(Appro: in out Comp; PP: out Comp; X,Y: in Scalar; Tay: in Taylor);--g_1
   procedure  ComputeTay(N: in Integer; Cont,FCont: in CompArray; Tay: out Taylor);
   procedure SetCont(Cont_Length,This_Machine,Num_Of_Machines: in Integer; First,Last: out Integer);



   ScalPi: constant Scalar:= Scal(Rep(3.14159_26535_89793_23),Rep(3.14159_26535_89793_24));
   ScalTwoPi: constant Scalar:=Two*ScalPi;
   ScalZero:  constant Scalar:=Scal(Zero);
   ScalOne:  constant Scalar:=Scal(One);
   ScalTwo:  constant Scalar:=Scal(Two);
   ScalTheta:  constant Scalar:=Scal(Rep(6.18033988749894848E-001));
   CoZero: constant Comp:=Complecs(Zero,Zero);
   CoOne: constant Comp:=Complecs(One,Zero);
   CoTwo: constant Comp:=Complecs(Two,Zero);
   CoTheta: constant Comp:=Complecs(Rep(6.18033988749894848E-001),Zero);
   CoI: constant Comp:=Complecs(Zero,One);
   CoPi: constant Comp:=SetBall(ScalPi,ScalZero);
   CoITheta: constant Comp:=Complecs(Zero,Rep(6.18033988749894848E-001));
   CoThetaSqr: constant Comp:=Sqr(CoTheta);
   CoTwoPi: constant Comp:=Two*CoPi;
   CoExpTheta: constant Comp:=Exp(Two*ScalPi*CoITheta);
   CrPoint: constant Comp:=(-Half)*CoExpTheta;
   EX: Integer:=0;
   LogB,LogBB,Cn2,Cn2n,P,TayCenter: Comp;
   Rho,RhoC: Rep;
   LogA: Scalar;
   Qn,Qnplus1,Qnplus2,Qnmin1,Qnmin2: Integer;
   Quad: Boolean;

private
   type Parametrization is record
      C: Comp;
      N: Integer;
   end record;

   Trunc: constant Boolean := (IsNumeric(ScalZero) and IsCoNumber(CoZero));

end RG;
