with Messages,Ada.IO_Exceptions;
use Messages;

package body Taylors.IO is

   pragma Suppress(Storage_Check);

   package Txt_Int_IO is new Ada.Text_IO.Integer_IO(Integer);
   use Txt_Int_IO;

   procedure Show(P: in Taylor; Cut: in Rep := Zero) is
      C: Comp;
      C1: Rep;
   begin
      for K in 0 .. Deg(P) loop
         C := Component(K,P);
         if SupMod(C)>Cut then
            New_Line;
            Put(Current_Output,"z^");
            Put(Current_Output,K,2,10);
            Put(Current_Output,"   ");
            Show(C);
         end if;
      end loop;
      C1 := ErrComp(P);
      if C1>Cut then
         New_Line;
         Put(Current_Output," E   ");
         Txt_Put(Current_Output,C1);
      end if;
      New_Line;
   end Show;


   procedure Get(R: in Radius; F: in File_Type; P: out Taylor; GetPoly: in Boolean;  Decimal: in Boolean := True) is
      D: Integer;
      E: Comp := CompZero;
      C: Rep;
      C1: Comp;
      B: Rep:= R;
   begin
      SetZero(P);
      Get(File=>F, Item=>D);
      P.D:=Size;
      if D<Size then
         P.D:=D;
      end if;
      if D<0 then return; end if;

      if not GetPoly then
         Txt_Get(F,C,Decimal);
      else
         C:=Zero;
      end if;
      SetErrComp(C,P);

      for K in 0 .. D loop
         Get(F,C1,Decimal, not GetPoly);
         if K <= Size then
            SetComponent(K,C1,P);
         else
            E := E+B*Ball0(SupMod(C1));
            B := R*B;
         end if;
      end loop;
      SetComponent(Size,Component(Size,P)+E,P);  -- add higher order error
   end Get;

   procedure Put(F: in File_Type; P: in Taylor; PutPoly: in Boolean; Decimal: in Boolean := True) is
      Pt: Taylor;
   begin
      Copy(P,Pt);
      Cleanup(Pt);
      Put(File=>F, Item=>Deg(Pt), Width=>3);
      New_Line(F);
      if Deg(Pt)<0 then return; end if;
      if not PutPoly then
         Txt_Put(F,ErrComp(Pt),Decimal);
         New_Line(F);
      end if;
      for K in 0 .. Deg(Pt) loop
         Put(F,Component(K,Pt),Decimal,not PutPoly);
         New_Line(F);
      end loop;
   end Put;

   procedure Read(File_Name: in String; R: in Radius; P: out Taylor;GetPoly: in Boolean;  Decimal: in Boolean := True) is
      F: File_Type;
   begin
      Open(F,In_File,File_Name);
      Get(R,F,P,GetPoly,Decimal);
      Close(F);
   end Read;

   procedure Read_Par(File_Name: in String; R: in Radius; P: out Taylor; GetPoly: in Boolean; Decimal: in Boolean := True) is
      F: File_Type;
   begin
      if EXX=1 then
         New_Line;
         Put("Handle is not reset");Message("CoNumbers.IO.Read_P",Parameter_Error);
      end if;
      Open(F,In_File,File_Name);
      Get(R,F,P,GetPoly,Decimal);
      Close(F);
   exception
      when Ada.Text_IO.End_Error =>
         Close(F);
         Put("handling exception...");
         EXX:=1;
         return;
      when Ada.IO_Exceptions.Data_Error =>
         Close(F);
         Put("handling exception...");
         EXX:=1;
         return;
      when Ada.IO_Exceptions.Name_Error =>
         Put("handling exception...");
         EXX:=1;
         return;
      when Ada.IO_Exceptions.Status_Error =>
         Put("handling exception...");
         EXX:=1;
         return;
      when Ada.IO_Exceptions.Use_Error =>
         Put("handling exception...");
         EXX:=1;
         return;
   end Read_Par;

   procedure Read_P(File_Name: in String; R: in Radius; P: out Taylor; GetPoly: in Boolean; Decimal: in Boolean := True) is
      F: File_Type;
   begin
      <<REREAD>> Read_Par(File_Name,R,P,GetPoly,Decimal);
      if EXX=1 then
         EXX:=0;
         delay 0.01;
         goto REREAD;
      end if;
   end Read_P;

   procedure Write(File_Name: in String; P: in Taylor; PutPoly: in Boolean; Decimal: in Boolean := True) is
      F: File_Type;
   begin
      Create(F,Out_File,File_Name);
      Put(F,P,PutPoly,Decimal);
      Close(F);
   end Write;
end Taylors.IO;
